﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

class MP4Box
{
    public string boxType;
    public int boxLen;
    public byte[] buff;
}

class Chunk
{
    public int offset;
    public int size;
    public int delta;
}

class MP4Reader
{
    Dictionary<string, object> dicHasChild;
    MP4Box stsz;
    MP4Box stts;
    MP4Box stco;
    bool gpsTrack = false;

    public List<GpsData> gpsRead(string filename)
    {
        List<GpsData> gpsDataList = new List<GpsData>();
        using (FileStream fs = new FileStream(filename, FileMode.Open, FileAccess.Read))
        {
            // 初期化
            dicHasChild = new Dictionary<string, object>();
            dicHasChild.Add("moov", null);
            dicHasChild.Add("trak", null);
            dicHasChild.Add("mdia", null);
            dicHasChild.Add("minf", null);
            dicHasChild.Add("stbl", null);
            gpsTrack = false;
            stsz = null;
            stts = null;
            stco = null;

            // moovの中のGPSトラックのst何とかを読みこむ
            MP4BoxReader mp4r = new MP4BoxReader(fs, fs.Length, dicHasChild);
            mp4readloop(mp4r, 0);

            if (stsz == null || stts == null || stco == null)
            {
                // st何とかが見つからなかったら残念。
                System.Diagnostics.Debugger.Break();
                return gpsDataList;
            }

            Array.Reverse(stsz.buff, 8, 4);
            int stszNum = BitConverter.ToInt32(stsz.buff, 8);
            Array.Reverse(stco.buff, 4, 4);
            int stcoNum = BitConverter.ToInt32(stco.buff, 4);
            Array.Reverse(stts.buff, 4, 4);
            int sttsNum = BitConverter.ToInt32(stts.buff, 4);
            if (stszNum != stcoNum)
            {
                // 数があってないと困る。
                System.Diagnostics.Debugger.Break();
                return gpsDataList;
            }

            // 扱いやすいように byte配列から Chunkの配列へ変換
            Chunk[] chunks = new Chunk[stcoNum];
            for (int i = 0, j = 0; i < stcoNum; i++)
            {
                chunks[i] = new Chunk();
                Array.Reverse(stco.buff, 8 + i * 4, 4);
                chunks[i].offset = BitConverter.ToInt32(stco.buff, 8 + i * 4);
                Array.Reverse(stsz.buff, 12 + i * 4, 4);
                chunks[i].size = BitConverter.ToInt32(stsz.buff, 12 + i * 4);
                //Array.Reverse(stsz.buff, 4 + j * 8, 4);
                //chunks[i].size = BitConverter.ToInt32(stsz.buff, 4 + j * 8);
            }

            // Chunk配列に従って、内容読み取り
            fs.Seek(0, SeekOrigin.Begin);
            Byte[] buff = new Byte[1024];
            for (int i = 0; i < stcoNum; i++)
            {
                fs.Seek(chunks[i].offset, SeekOrigin.Begin);
                fs.Read(buff, 0, chunks[i].size);

                if (buff[0] == 'G')
                {
                    // GPSデータ
                    string nmearmc = Encoding.ASCII.GetString(buff, 0, chunks[i].size);
                    //System.Diagnostics.Debug.Print(nmearmc);
                    GpsData gpsData = new GpsData(nmearmc);
                    gpsDataList.Add(gpsData);
                }
                /*
                else if (buff[0] == 'M')
                {
                    // Gセンサーデータ
                    int year = buff[7] + buff[8] * 256;
                    int month = buff[9];
                    int mday = buff[10];
                    int hour = buff[11];
                    int min = buff[12];
                    int sec = buff[13];
                    //string data = Encoding.ASCII.GetString(buff, 0, chunks[i].size);
                    System.Diagnostics.Debug.Print("{0:x2} {1:x2} {2:x2}  {3:x2} {4:x2} {5:x2} {6:0000}-{7:00}-{8:00} {9:00}:{10:00}:{11:00} {12:x2}",
                        buff[1], buff[2], buff[3], buff[4], buff[5], buff[6],
                        year, month, mday, hour, min, sec, buff[14]);
                }
                */
            }
        }
        return gpsDataList;
    }

    private void mp4readloop(MP4BoxReader mp4r, int level)
    {
        string indent = new String(' ', level * 2); // デバッグ出力用

        while (!mp4r.isEOB())
        {
            MP4Box mp4Box = mp4r.read();

            if (mp4Box.boxType == "hdlr")
            {
                string s1 = Encoding.ASCII.GetString(mp4Box.buff, 8, 4);
                string s2 = Encoding.ASCII.GetString(mp4Box.buff, 24, 4);

                // GPSトラックかどうかの確認。（これでいいのかなぁ？）
                gpsTrack = (s1 == "sdsm" && s2 == "XBLB");
            }
            else if (dicHasChild.ContainsKey(mp4Box.boxType))
            {
                // 子ブロックがある場合はそれを再帰的に読む。
                MP4BoxReader childBox = new MP4BoxReader(mp4r.fs, mp4Box.boxLen - 8, dicHasChild);
                mp4readloop(childBox, level + 1);
            }
            else if (gpsTrack)
            {
                switch (mp4Box.boxType)
                {
                    case "stsz":
                        stsz = mp4Box;
                        break;
                    case "stts":
                        stts = mp4Box;
                        break;
                    case "stco":
                        stco = mp4Box;
                        break;
                }
            }
            /*
            else
            {
                if (16 <= mp4Box.boxLen)
                {
                    if ('a' <= mp4Box.buff[4] && mp4Box.buff[4] <= 'z' &&
                        'a' <= mp4Box.buff[5] && mp4Box.buff[5] <= 'z' &&
                        'a' <= mp4Box.buff[6] && mp4Box.buff[6] <= 'z' &&
                        'a' <= mp4Box.buff[7] && mp4Box.buff[7] <= 'z')
                    {
                        Array.Reverse(mp4Box.buff, 0, 4);
                        int boxLen = BitConverter.ToInt32(mp4Box.buff, 0);
                        string boxType = Encoding.ASCII.GetString(mp4Box.buff, 4, 4);
                        System.Diagnostics.Debug.Print(indent + "  " + boxType + "-" + boxLen);
                    }
                }
            }
            */
        }
    }
}

class MP4BoxReader
{
    public FileStream fs = null;

    private byte[] buff;     // データ格納用配列
    private long remain;     // 読み込むべき残りのバイト数

    private Dictionary<string, object> dicHasChild;

    public MP4BoxReader(FileStream fileStream, long length, Dictionary<string, object> dicHasChild)
    {
        this.fs = fileStream;
        this.remain = length; // 対象boxのサイズ
        this.dicHasChild = dicHasChild;

        this.buff = new byte[8];
    }

    public bool isEOB()
    {
        // End of Box か？
        return remain <= 0;
    }

    public MP4Box read()
    {
        MP4Box mp4Box = new MP4Box();

        fs.Read(buff, 0, 8);

        Array.Reverse(buff, 0, 4);
        mp4Box.boxLen = BitConverter.ToInt32(buff, 0);
        mp4Box.boxType = Encoding.ASCII.GetString(buff, 4, 4);

        if (!dicHasChild.ContainsKey(mp4Box.boxType))
        {
            // 子ノードが無い場合は読み進める。

            if (mp4Box.boxType == "mdat")
            {
                // mdatはデカイので、読まずにスキップ
                fs.Seek(mp4Box.boxLen - 8, SeekOrigin.Current);
            }
            else
            {
                // 他はまあ小さいので、あまり考えずに読む。
                mp4Box.buff = new byte[mp4Box.boxLen - 8];
                fs.Read(mp4Box.buff, 0, mp4Box.boxLen - 8);
            }
        }
        remain -= mp4Box.boxLen;

        return mp4Box;
    }
}
